#############################################################################
# womble_plugin.tcl Copyright 2006, 2007, 2008, 2009 r0lZ
# PgcEdit plugin to create .MBS and .MSL files for Womble MPEG Video Wizard DVD.
#
# This plugin is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This plugin is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#############################################################################
# Usage: copy this file in the <PgcEdit's install dir>/plugins directory.
# A "Womble" menu will be added in the Plugins menu.
# You need PgcEdit 0.6.0 or more to use this plugin.
#############################################################################
# Version History:
#
# v 1.2 (June 17, 2009)
# - Added Create Celltimes.txt from a Womble PAL Project file
# v 1.1 (March 2, 2007)
# - Fixed a bug in the generation of the default file name
# v 1.0 (December 22, 2006)
# - First public release
#############################################################################


# Add Plugins -> Womble submenu
if {! [winfo exists .mainmenu.plugins.womble]} {
	.mainmenu.plugins add cascade -label "Womble" -menu .mainmenu.plugins.womble
	menu .mainmenu.plugins.womble
	.mainmenu.plugins.womble add command -label "Create a MBS file for current Title domain" \
			-command {::plugins::womble::create_mbs_domain}
	.mainmenu.plugins.womble add command -label "Create a MBS file for each Title domain in the DVD" \
			-command {::plugins::womble::create_mbs_all_domains}
	.mainmenu.plugins.womble add command -label "Create a MSL chapter list for current domain" \
			-command {::plugins::womble::build_cell_times_domain}
	.mainmenu.plugins.womble add separator
	.mainmenu.plugins.womble add command -label "Create Celltimes.txt from a Womble PAL Project file" \
			-command {::plugins::womble::wbp2celltimes}
	if {$::tcl_platform(platform) == "windows"} {
		.mainmenu.plugins.womble add separator
		.mainmenu.plugins.womble add command -label "Explore last output folder" \
				-command {::plugins::womble::explore_output_folder}
	}
	.mainmenu.plugins.womble add separator
	.mainmenu.plugins.womble add command -label "About" -command {
			tk_messageBox  -title "Womble plugin" -icon info \
					-message "Womble plugin for PgcEdit 0.6.0 or more.\nVersion $::plugins::womble::version\nAuthor: r0lZ\nFreeware!"
		}
}

namespace eval ::plugins {}
namespace eval ::plugins::womble {

set version "1.2"

set default_out_dir ""

################### MBS (Files list) functions ##################

proc create_mbs_domain {} {
	if {![info exists ::dvddir]} {bell; return}
	set index [lindex $::editor_cmds_from 1]
	set vts [lindex [split $index ,] 0]
	if {$vts < 1} {bell; return}
	create_mbs_vts $vts
}

proc create_mbs_all_domains {} {
	if {![info exists ::dvddir]} {bell; return}
	for {set vts 1} {$vts <= $::pgcs(numvts)} {incr vts} {
		create_mbs_vts $vts
	}
}

proc create_mbs_vts {vts {quiet false}} {
	show_msg "Creating MBS file for the VTS $vts Title domain..."
	set numfiles 0
	set txt ""
	for {set i 1} {$i < 10} {incr i} {
		set fname [file nativename [file normalize [file join $::dvddir [format VTS_%02d_%d.VOB $vts $i]]]]
		if {[file exists $fname]} {
			incr numfiles
			set txt "${txt}File=$fname\n"
		} else {
			break
		}
	}
	if {$numfiles == 0} {
		tk_messageBox -title "Womble plugin" -icon error -message \
				"No VOB file found for the title domain of VTS $vts."
		return
	}
	set out [string trim "\[MBSFile\]\nNumFile=$numfiles\n$txt"]
	if {$::plugins::womble::default_out_dir == ""} {
		set defaultdir [file dirname $::dvddir]
	} else {
		set defaultdir $::plugins::womble::default_out_dir
	}
	set defaultfile [default_filename pgc $vts]
	set outname [tk_getSaveFile -filetypes {{"MBS files" {*.mbs}}} \
			-initialdir $defaultdir -initialfile ${defaultfile}.mbs \
			-parent . -title "Save Womble MBS file as..."]
	if {$outname != ""} {
		set ::plugins::womble::default_out_dir [file dirname $outname]
		if {[string toupper [file extension $outname]] != ".MBS"} {
			set outname $outname.mbs
		}
		saveFile $outname auto $out
		show_msg "MBS file (VOB files list) saved OK."

		# create also a frame list?
		set txt "MBS file saved OK:\n\"$outname\""
		set txt "$txt\n\nDo you want to save also a bookmark list (chapters) for this domain?"
		set txt "$txt\nA bookmark will be created for each cell in the VOB file(s)."
		set txt "$txt\n(The file will be saved with the same name but with a .MLS extension.)"
		set txt "$txt\n\nThis operation can take some time.  Please be patient!"
		set r [tk_messageBox -title "Womble plugin" -icon question -message $txt \
				-type yesno -default yes]
		if {$r == "yes"} {
			set framelist [build_cell_times pgc $vts]
			set txt [cell_times_to_text pgc $vts $framelist $outname]
			saveFile [string range $outname 0 [string last . $outname]].mls auto $txt
			show_msg "Bookmark file saved OK."
		}
	}
}

################### MLS (Chapters list) functions ##################

proc build_cell_times_domain {} {
	if {![info exists ::dvddir]} {bell; return}
	set type  [lindex $::editor_cmds_from 0]
	set index [lindex $::editor_cmds_from 1]
	set vts [lindex [split $index ,] 0]
	if {$type == "pgc" && $vts < 1} {bell; return}
	set framelist [build_cell_times $type $vts]
	set txt [cell_times_to_text $type $vts $framelist]

	if {$::plugins::womble::default_out_dir == ""} {
		set defaultdir [file dirname $::dvddir]
	} else {
		set defaultdir $::plugins::womble::default_out_dir
	}
	set defaultfile [default_filename $type $vts]
	::main::showtext .mlsfile "MLS file.  Edit if necessary, and save it!" $txt \
			-defaultfilename [file join $defaultdir $defaultfile.mls]
	return
}

proc build_cell_times {type vts} {
	if {$vts == 0} {
		show_msg "Creating MLS file for the VMGM Menu domain..."
	} else {
		if {$type == "pgc"} {set domain Title} else {set domain Menu}
		show_msg "Creating MLS file for the VTS $vts $type domain..."
	}
	# display error message if VOB is desynchronised
	unset -nocomplain ::menued::load_nav_pack_error_skip

	# determine frame rate and file size
	if {$type == "menu"} {
		if {$vts == 0} {
			set vobfile [file join $::dvddir VIDEO_TS.VOB]
		} else {
			set vobfile [file join $::dvddir [format VTS_%02d_0.VOB $vts]]
		}
		set totalsize [expr [file size $vobfile] / 2048]
		set tmp [get_int ::menus($vts,1,1,header) 7 1]
	} else {
		set totalsize 0
		for {set i 1} {$i < 10} {incr i} {
			set vobfile [file join $::dvddir [format VTS_%02d_%d.VOB $vts $i]]
			if {[file exists $vobfile]} {
				incr totalsize [expr [file size $vobfile] / 2048]
			}
		}
		set vobfile [file join $::dvddir [format VTS_%02d_1.VOB $vts]]
		set tmp [get_int ::pgcs($vts,1,header) 7 1]
	}
	if {[expr $tmp & 0xC0] == 192} {
		set fps 30
	} else {
		set fps 25
	}

	set numcells 0
	set start_ptms {}
	set end_ptms {}
	set oldvid -1
	set oldcid -1
	set lba 0
	set sector 0
	set titlefilenum 1
	set refresh_progress 0

	set vobsize [expr [file size $vobfile] / 2048]
	set fh [::menued::openbinfile $vobfile r]
	set navpack [::menued::load_nav_pack $vobfile $fh $sector]
	if {$navpack==""} {
		close $fh
#		vob_warning $type $vts $vobfile $lba $sector "First Nav Pack is not valid"
		return ""
	}

	while 1 {
		incr refresh_progress
		if {$refresh_progress >= 100} {
			set refresh_progress 0
			update_progress $lba $totalsize
		}
		set vid [get_int navpack 1055 2]
		set cid [get_int navpack 1058 1]
		if {$vid != $oldvid || $cid != $oldcid} {
			incr numcells

			# compute new time offset for current cell
			set startptm [get_int navpack 57 4]
			if {$start_ptms != {}} {
				#tolog "End PTM: $lastvobuendptm"
				lappend end_ptms $lastvobuendptm
			}
			show_msg "Found cell $numcells, V/CID $vid/$cid, at LBA $lba (sector $sector in [file tail $vobfile])" $lba $totalsize
			lappend start_ptms $startptm
			#tolog "Start PTM: $startptm"
			set oldvid $vid
			set oldcid $cid
		}
		set lastvobuendptm [get_int navpack 61 4]

		set lbaoffset [get_int navpack 1039 4]
		set lba [expr {$lba+$lbaoffset+1}]
		set sector [expr {$sector+$lbaoffset+1}]
		if {$sector >= $vobsize || $sector==0} {
			close $fh
			if {$type == "pgc"} {
				set sector [expr $sector - $vobsize]
				incr titlefilenum
				set vobfile [file join $::dvddir [format VTS_%02d_%d.VOB $vts $titlefilenum]]
				if {[file exists $vobfile]} {
					set vobsize [expr [file size $vobfile] / 2048]
					set fh [::menued::openbinfile $vobfile r]
				} else {
					break
				}
			} else {
				break
			}
		}
		set navpack [::menued::load_nav_pack $vobfile $fh $sector]
		if {$navpack==""} {
			close $fh
#			vob_warning $type $vts $vobfile $lba $sector "Nav Pack is not valid"
			return ""
		}
	}
	lappend end_ptms $lastvobuendptm

	show_msg "Done.  Found $numcells cells." 0 0
	#tolog "Start PTMs: $start_ptms"
	#tolog "End   PTMs: $end_ptms"

	set framelist {}
	for {set i 0} {$i < $numcells} {incr i} {
		set numframes [clock90KHz2frames [expr [lindex $end_ptms $i]-[lindex $start_ptms $i]] $fps]
		lappend framelist $numframes
		#tolog "Cell [expr $i+1]: $numframes frames."
	}
	return $framelist
}

proc cell_times_to_text {type vts framelist {mbsfilename ""}} {
	set txt ""
	if {$mbsfilename == ""} {
		if {$vts == 0} {
			set filename [file join $::dvddir VIDEO_TS.VOB]
		} else {
			if {$type == "menu"} {
				set filename [file join $::dvddir [format VTS_%02d_0.VOB $vts]]
			} else {
				set filename [file join $::dvddir [format VTS_%02d_1.VOB $vts]]
				set txt "// ***\n// *** WARNING: YOU HAVE PROBABLY TO CHANGE THE PATH NAME BELOW!\n// ***\n"
			}
		}
	} else {
		set filename $mbsfilename
	}

	set len [llength $framelist]
	set txt "${txt}\[BookmarkList\]"
	set txt "$txt\nPathName= [file nativename [file normalize $filename]]"
	set txt "$txt\nVideoStreamID= 0"
	set txt "$txt\nFormat= frame"
	set txt "$txt\nCount= $len"
	set total 0
	for {set i 0} {$i < $len} {incr i} {
		set txt "$txt\n[format %12d $total] 1"
		incr total [lindex $framelist $i]
	}
	return $txt
}

proc clock90KHz2frames {clock fps} {
	if {$fps == 25} {
		set clkrate 90000
	} else {
		set clkrate 90090
	}
	return [expr $clock / ($clkrate/$fps)]
}

proc wbp2celltimes {} {
	set fps 25.0
	set ticks {}
	if {[info exists ::dvddir]} {
		set initialdir [file dirname $::dvddir]
	} else {
		set initialdir ""
	}
	set infile [tk_getOpenFile \
		-filetypes {{"Womble projects" {*.wbp}}} \
		-initialdir $initialdir \
		-title "Open source Womble project file" -parent .]
	if {$infile == ""} return
	set in [open $infile r]
	if {$in == ""} return
	fconfigure $in -translation auto
	while {[eof $in] == 0} {
		set line [gets $in]
		if {[string first "MarkerPosition" $line] == 0} {
			set idx [expr [string first "=" $line] + 1]
			set line [string range $line $idx end]
			set ticks "$ticks$line "
		}
		if {[lindex $line 0] == "VideoInfo="} {
			set fps [string trim [lindex $line 3] ,]
		}
	}
	close $in
	set ticks [string trim $ticks]
	if {[llength $ticks] == 0} return
	set out ""
	foreach i $ticks {
		set framenum [expr round($i * $fps / 30000)]
		set out "$out$framenum\n"
	}
	::main::showtext .celltimes "CellTimes.txt file from Womble project [file tail $infile]" $out \
			-defaultfilename [file join [file dirname $infile] "CellTimes.txt"]
	return
}

################### Misc functions ##################

proc default_filename {type vts} {
	set defaultfile [::utils::get_General_Name]
	if {$defaultfile == ""} {
		if {[file tail $::dvddir] != "VIDEO_TS"} {
			set defaultfile [file tail $::dvddir]
		} else {
			set defaultfile [file tail [file dirname $::dvddir]]
		}
		set defaultfile [string trim $defaultfile ":"]
	} else {
		set defaultfile [string map {: _ / _ \\\\ _ ? _ * _ \" _ | _ < _ > _ __ _} $defaultfile]
	}
	if {$vts == 0} {
		set defaultfile ${defaultfile}__VMGM
	} else {
		if {$type == "menu"} {
			set defaultfile ${defaultfile}__VTSM_[format %02d $vts]
		} else {
			set defaultfile ${defaultfile}__VTST_[format %02d $vts]
		}
	}
	return $defaultfile
}

proc explore_output_folder {} {
	if {$::tcl_platform(platform) != "windows"} {bell; return}
	if {$::plugins::womble::default_out_dir == ""} {bell; return}
	exec -- explorer.exe [file nativename $::plugins::womble::default_out_dir] &
}



}	;# end of namespace ::plugins::womble
